/**
* @license
* Copyright (C) 2017-2018 "IoT.bzh"
* Author Sebastien Douheret <sebastien@iot.bzh>
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*   http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

import { Injectable } from '@angular/core';
import { CookieService } from 'ngx-cookie';
import { NbThemeService } from '@nebular/theme';

import { Observable } from 'rxjs/Observable';
import { BehaviorSubject } from 'rxjs/BehaviorSubject';

export interface IConfig {
  language: string;
  theme: string;
  grafanaDashboardUrl: string;
}

@Injectable()
export class ConfigService {

  public Conf$: Observable<IConfig>;

  private confSubject: BehaviorSubject<IConfig>;
  private confStore: IConfig;

  private confDefault: IConfig = {
    language: 'ENG',
    theme: 'default',
    grafanaDashboardUrl: 'http://localhost:3000',
  };

  constructor(
    private cookie: CookieService,
    private themeService: NbThemeService,
  ) {
    this.confSubject = <BehaviorSubject<IConfig>>new BehaviorSubject(this.confStore);
    this.Conf$ = this.confSubject.asObservable();

    // Save selected theme in cookie
    this.themeService.onThemeChange().subscribe(tm => {
      if (typeof this.confStore === 'undefined') {
        return;
      }
      if (tm.name !== this.confStore.theme) {
        this.confStore.theme = tm.name;
        this.save(this.confStore);
      }
    });

    // Load initial config and apply it
    this.load();
    this.themeService.changeTheme(this.confStore.theme);
  }

  // Load config
  load() {
    // Try to retrieve previous config from cookie
    const cookConf = this.cookie.getObject('xds-config');
    if (cookConf != null) {
      this.confStore = <IConfig>cookConf;
      this.confSubject.next(Object.assign({}, this.confStore));
    } else {
      // Set default config
      this.save(null);
    }
  }

  // Save config into cookie
  save(cfg: IConfig | null) {
    this.confStore = (cfg !== null) ? cfg : this.confDefault;

    // Notify subscribers
    this.confSubject.next(Object.assign({}, this.confStore));

    this.cookie.putObject('xds-config', this.confStore);
  }

}
